@@ -22,7 +22,7 @@
 
 /* This file contains common stuff in Ephemeral Diffie-Hellman (DHE)
  * and Anonymous DH key exchange(DHA). These are used in the handshake
- * procedure of the certificate and anoymous authentication.
+ * procedure of the certificate and anonymous authentication.
  */
 
 #include "gnutls_int.h"
@@ -228,6 +228,7 @@ _gnutls_proc_dh_common_server_kx (gnutls_session_t session,
 
   DECR_LEN (data_size, n_Y);
   data_Y = &data[i];
+  i += n_Y;
 
   _n_Y = n_Y;
   _n_g = n_g;
@@ -296,27 +297,26 @@ _gnutls_dh_common_print_server_kx (gnutls_session_t session,
   ret = _gnutls_buffer_append_mpi(data, 16, p, 0);
   if (ret < 0)
     {
-      gnutls_assert();
+      ret = gnutls_assert_val(ret);
       goto cleanup;
     }
 
   ret = _gnutls_buffer_append_mpi(data, 16, g, 0);
   if (ret < 0)
     {
-      gnutls_assert();
+      ret = gnutls_assert_val(ret);
       goto cleanup;
     }
 
   ret = _gnutls_buffer_append_mpi(data, 16, Y, 0);
   if (ret < 0)
     {
-      gnutls_assert();
+      ret = gnutls_assert_val(ret);
       goto cleanup;
     }
-  
-  ret = data->length;
+
 cleanup:
   _gnutls_mpi_release (&Y);
 
-  return ret;
+  return data->length;
 }
