@@ -155,9 +155,26 @@ int ret;
   return ret;
 }
 
+inline static 
+int errno_to_gerr(int err)
+{
+  switch(err)
+    {
+      case EAGAIN:
+        return GNUTLS_E_AGAIN;
+      case EINTR:
+        return GNUTLS_E_INTERRUPTED;
+      case EMSGSIZE:
+        return GNUTLS_E_LARGE_PACKET;
+      default:
+        gnutls_assert ();
+        return GNUTLS_E_PUSH_ERROR;
+    }
+}
+
 static ssize_t
 _gnutls_dgram_read (gnutls_session_t session, mbuffer_st **bufel,
-		    gnutls_pull_func pull_func)
+		    gnutls_pull_func pull_func, unsigned int ms)
 {
   ssize_t i, ret;
   uint8_t *ptr;
@@ -168,14 +185,21 @@ _gnutls_dgram_read (gnutls_session_t session, mbuffer_st **bufel,
   if (recv_size > max_size)
     recv_size = max_size;
 
+  session->internals.direction = 0;
+
+  if (ms)
+    {
+      ret = _gnutls_io_check_recv(session, ms);
+      if (ret < 0)
+        return gnutls_assert_val(ret);
+    }
+
   *bufel = _mbuffer_alloc (0, max_size);
   if (*bufel == NULL)
     return gnutls_assert_val(GNUTLS_E_MEMORY_ERROR);
 
   ptr = (*bufel)->msg.data;
-
-  session->internals.direction = 0;
-
+  
   reset_errno (session);
   i = pull_func (fd, ptr, recv_size);
 
@@ -186,22 +210,8 @@ _gnutls_dgram_read (gnutls_session_t session, mbuffer_st **bufel,
       _gnutls_read_log ("READ: %d returned from %p, errno=%d gerrno=%d\n",
 			(int) i, fd, errno, session->internals.errnum);
 
-      if (err == EAGAIN)
-        {
-          ret = GNUTLS_E_AGAIN;
-          goto cleanup;
-        }
-      else if (err == EINTR)
-        {
-          ret = GNUTLS_E_INTERRUPTED;
-          goto cleanup;
-        }
-      else
-        {
-          gnutls_assert ();
-          ret = GNUTLS_E_PULL_ERROR;
-          goto cleanup;
-        }
+      ret = errno_to_gerr(err);
+      goto cleanup;
     }
   else
     {
@@ -229,13 +239,18 @@ cleanup:
 
 static ssize_t
 _gnutls_stream_read (gnutls_session_t session, mbuffer_st **bufel,
-		     size_t size, gnutls_pull_func pull_func)
+		     size_t size, gnutls_pull_func pull_func, unsigned int ms)
 {
   size_t left;
   ssize_t i = 0;
   size_t max_size = _gnutls_get_max_decrypted_data(session);
   uint8_t *ptr;
   gnutls_transport_ptr_t fd = session->internals.transport_recv_ptr;
+  int ret;
+  struct timespec t1, t2;
+  unsigned int diff;
+
+  session->internals.direction = 0;
 
   *bufel = _mbuffer_alloc (0, MAX(max_size, size));
   if (!*bufel)
@@ -245,11 +260,18 @@ _gnutls_stream_read (gnutls_session_t session, mbuffer_st **bufel,
     }
   ptr = (*bufel)->msg.data;
 
-  session->internals.direction = 0;
-
   left = size;
   while (left > 0)
     {
+      if (ms)
+        {
+          ret = _gnutls_io_check_recv(session, ms);
+          if (ret < 0)
+            return gnutls_assert_val(ret);
+            
+          gettime(&t1);
+        }
+
       reset_errno (session);
 
       i = pull_func (fd, &ptr[size - left], left);
@@ -272,9 +294,7 @@ _gnutls_stream_read (gnutls_session_t session, mbuffer_st **bufel,
                   goto finish;
                 }
 
-              if (err == EAGAIN)
-                return GNUTLS_E_AGAIN;
-              return GNUTLS_E_INTERRUPTED;
+              return errno_to_gerr(err);
             }
           else
             {
@@ -290,9 +310,19 @@ _gnutls_stream_read (gnutls_session_t session, mbuffer_st **bufel,
           if (i == 0)
             break;              /* EOF */
         }
-
+        
       left -= i;
       (*bufel)->msg.size += i;
+
+      if (ms && left)
+        {
+          gettime(&t2);
+          diff = _dtls_timespec_sub_ms(&t2, &t1);
+          if (diff < ms)
+            ms -= diff;
+          else
+            return gnutls_assert_val(GNUTLS_E_TIMEDOUT);
+        }
     }
 
 finish:
@@ -311,13 +341,13 @@ finish:
  */
 static ssize_t
 _gnutls_read (gnutls_session_t session, mbuffer_st **bufel,
-	      size_t size, gnutls_pull_func pull_func)
+	      size_t size, gnutls_pull_func pull_func, unsigned int ms)
 {
   if (IS_DTLS (session))
     /* Size is not passed, since a whole datagram will be read. */
-    return _gnutls_dgram_read (session, bufel, pull_func);
+    return _gnutls_dgram_read (session, bufel, pull_func, ms);
   else
-    return _gnutls_stream_read (session, bufel, size, pull_func);
+    return _gnutls_stream_read (session, bufel, size, pull_func, ms);
 }
 
 static ssize_t
@@ -347,6 +377,7 @@ _gnutls_writev_emu (gnutls_session_t session, gnutls_transport_ptr_t fd, const g
   return ret;
 }
 
+
 static ssize_t
 _gnutls_writev (gnutls_session_t session, const giovec_t * giovec,
                 int giovec_cnt)
@@ -365,15 +396,8 @@ _gnutls_writev (gnutls_session_t session, const giovec_t * giovec,
     {
       int err = get_errno (session);
       _gnutls_debug_log ("errno: %d\n", err);
-      if (err == EAGAIN)
-        return GNUTLS_E_AGAIN;
-      else if (err == EINTR)
-        return GNUTLS_E_INTERRUPTED;
-      else
-        {
-          gnutls_assert ();
-          return GNUTLS_E_PUSH_ERROR;
-        }
+
+      return errno_to_gerr(err);
     }
   return i;
 }
@@ -386,10 +410,13 @@ _gnutls_writev (gnutls_session_t session, const giovec_t * giovec,
  * This is not a general purpose function. It returns EXACTLY the data requested,
  * which are stored in a local (in the session) buffer.
  *
+ * If the @ms parameter is non zero then this function will return before
+ * the given amount of milliseconds or return GNUTLS_E_TIMEDOUT.
+ *
  */
 ssize_t
 _gnutls_io_read_buffered (gnutls_session_t session, size_t total,
-                          content_type_t recv_type)
+                          content_type_t recv_type, unsigned int ms)
 {
   ssize_t ret = 0;
   size_t min;
@@ -445,7 +472,7 @@ _gnutls_io_read_buffered (gnutls_session_t session, size_t total,
     {
       ret =
         _gnutls_read (session, &bufel, readsize,
-                      session->internals.pull_func);
+                      session->internals.pull_func, ms);
 
       /* return immediately if we got an interrupt or eagain
        * error.
@@ -591,6 +618,13 @@ _gnutls_io_write_flush (gnutls_session_t session)
                          (int) send_buffer->byte_length);
       return ret;
     }
+  else if (ret == GNUTLS_E_LARGE_PACKET)
+    {
+      _mbuffer_head_remove_bytes (send_buffer, tosend);
+      _gnutls_write_log ("WRITE cannot send large packet (%u bytes).\n",
+                         (unsigned int) tosend);
+      return ret;
+    }
   else
     {
       _gnutls_write_log ("WRITE error: code %d, %d bytes left.\n",
@@ -763,7 +797,7 @@ parse_handshake_header (gnutls_session_t session, mbuffer_st* bufel,
   dataptr = _mbuffer_get_udata_ptr(bufel);
 
   /* if reading a client hello of SSLv2 */
-  if (!IS_DTLS(session) && bufel->htype == GNUTLS_HANDSHAKE_CLIENT_HELLO_V2)
+  if (unlikely(!IS_DTLS(session) && bufel->htype == GNUTLS_HANDSHAKE_CLIENT_HELLO_V2))
     {
       handshake_header_size = SSL2_HEADERS; /* we've already read one byte */
 
@@ -778,7 +812,7 @@ parse_handshake_header (gnutls_session_t session, mbuffer_st* bufel,
       hsk->start_offset = 0;
       hsk->end_offset = hsk->length;
     }
-  else /* TLS handshake headers */
+  else /* TLS or DTLS handshake headers */
     {
 
       hsk->htype = dataptr[0];
@@ -799,7 +833,7 @@ parse_handshake_header (gnutls_session_t session, mbuffer_st* bufel,
         {
           hsk->sequence = 0;
           hsk->start_offset = 0;
-          hsk->end_offset = hsk->length;
+          hsk->end_offset = MIN((_mbuffer_get_udata_size(bufel) - handshake_header_size), hsk->length);
         }
     }
   data_size = _mbuffer_get_udata_size(bufel) - handshake_header_size;
@@ -1143,6 +1177,7 @@ _gnutls_handshake_io_recv_int (gnutls_session_t session,
                                handshake_buffer_st * hsk, unsigned int optional)
 {
   int ret;
+  time_t tleft = 0;
 
   ret = get_last_packet(session, htype, hsk, optional);
   if (ret != GNUTLS_E_AGAIN && ret != GNUTLS_E_INTERRUPTED && ret != GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE)
