@@ -42,8 +42,8 @@
 #include "x509/x509_int.h"
 #include <gnutls_str_array.h>
 #include "read-file.h"
-#if defined _WIN32 || defined __WIN32__
-#include <wincrypt.h>
+#ifdef _WIN32
+# include <wincrypt.h>
 #endif
 
 /*
@@ -478,51 +478,6 @@ read_cert_mem (gnutls_certificate_credentials_t res, const void *cert,
   return ret;
 }
 
-static int
-_gnutls_x509_raw_privkey_to_privkey (gnutls_privkey_t * privkey,
-                                     const gnutls_datum_t * raw_key,
-                                     gnutls_x509_crt_fmt_t type)
-{
-  gnutls_x509_privkey_t tmpkey;
-  int ret;
-
-  ret = gnutls_x509_privkey_init (&tmpkey);
-  if (ret < 0)
-    {
-      gnutls_assert ();
-      return ret;
-    }
-
-  ret = gnutls_x509_privkey_import (tmpkey, raw_key, type);
-  if (ret < 0)
-    {
-      gnutls_assert ();
-      gnutls_x509_privkey_deinit (tmpkey);
-      return ret;
-    }
-
-  ret = gnutls_privkey_init (privkey);
-  if (ret < 0)
-    {
-      gnutls_assert ();
-      gnutls_x509_privkey_deinit (tmpkey);
-      return ret;
-    }
-
-  ret =
-    gnutls_privkey_import_x509 (*privkey, tmpkey,
-                                GNUTLS_PRIVKEY_IMPORT_AUTO_RELEASE);
-  if (ret < 0)
-    {
-      gnutls_assert ();
-      gnutls_x509_privkey_deinit (tmpkey);
-      gnutls_privkey_deinit (*privkey);
-      return ret;
-    }
-
-  return 0;
-}
-
 /* Reads a PEM encoded PKCS-1 RSA/DSA private key from memory.  Type
  * indicates the certificate format.  KEY can be NULL, to indicate
  * that GnuTLS doesn't know the private key.
@@ -540,7 +495,17 @@ read_key_mem (gnutls_certificate_credentials_t res,
       tmp.data = (uint8_t *) key;
       tmp.size = key_size;
 
-      ret = _gnutls_x509_raw_privkey_to_privkey (&privkey, &tmp, type);
+      ret = gnutls_privkey_init(&privkey);
+      if (ret < 0)
+        {
+          gnutls_assert ();
+          return ret;
+        }
+      
+      if (res->pin.cb)
+        gnutls_privkey_set_pin_function(privkey, res->pin.cb, res->pin.data);
+
+      ret = gnutls_privkey_import_x509_raw (privkey, &tmp, type, NULL, 0);
       if (ret < 0)
         {
           gnutls_assert ();
@@ -566,43 +531,28 @@ read_key_mem (gnutls_certificate_credentials_t res,
   return 0;
 }
 
-#ifdef ENABLE_PKCS11
+
 /* Reads a private key from a token.
  */
 static int
 read_key_url (gnutls_certificate_credentials_t res, const char *url)
 {
   int ret;
-  gnutls_pkcs11_privkey_t key1 = NULL;
   gnutls_privkey_t pkey = NULL;
 
   /* allocate space for the pkey list
    */
-
-  ret = gnutls_pkcs11_privkey_init (&key1);
+  ret = gnutls_privkey_init (&pkey);
   if (ret < 0)
     {
       gnutls_assert ();
       return ret;
     }
 
-  ret = gnutls_pkcs11_privkey_import_url (key1, url, 0);
-  if (ret < 0)
-    {
-      gnutls_assert ();
-      goto cleanup;
-    }
+  if (res->pin.cb)
+    gnutls_privkey_set_pin_function(pkey, res->pin.cb, res->pin.data);
 
-  ret = gnutls_privkey_init (&pkey);
-  if (ret < 0)
-    {
-      gnutls_assert ();
-      goto cleanup;
-    }
-
-  ret =
-    gnutls_privkey_import_pkcs11 (pkey, key1,
-                                  GNUTLS_PRIVKEY_IMPORT_AUTO_RELEASE);
+  ret = gnutls_privkey_import_url (pkey, url, 0);
   if (ret < 0)
     {
       gnutls_assert ();
@@ -622,12 +572,10 @@ cleanup:
   if (pkey)
     gnutls_privkey_deinit (pkey);
 
-  if (key1)
-    gnutls_pkcs11_privkey_deinit (key1);
-
   return ret;
 }
 
+#ifdef ENABLE_PKCS11
 /* Reads a private key from a token.
  */
 static int
@@ -641,7 +589,7 @@ read_cas_url (gnutls_certificate_credentials_t res, const char *url)
   /* FIXME: should we use login? */
   ret =
     gnutls_pkcs11_obj_list_import_url (NULL, &pcrt_list_size, url,
-                                       GNUTLS_PKCS11_OBJ_ATTR_CRT_TRUSTED_CA, 0);
+                                       GNUTLS_PKCS11_OBJ_ATTR_CRT_TRUSTED, 0);
   if (ret < 0 && ret != GNUTLS_E_SHORT_MEMORY_BUFFER)
     {
       gnutls_assert ();
@@ -663,7 +611,7 @@ read_cas_url (gnutls_certificate_credentials_t res, const char *url)
 
   ret =
     gnutls_pkcs11_obj_list_import_url (pcrt_list, &pcrt_list_size, url,
-                                       GNUTLS_PKCS11_OBJ_ATTR_CRT_TRUSTED_CA, 0);
+                                       GNUTLS_PKCS11_OBJ_ATTR_CRT_TRUSTED, 0);
   if (ret < 0)
     {
       gnutls_assert ();
@@ -681,9 +629,10 @@ read_cas_url (gnutls_certificate_credentials_t res, const char *url)
   ret =
     gnutls_x509_crt_list_import_pkcs11 (xcrt_list, pcrt_list_size, pcrt_list,
                                         0);
-  if (ret < 0)
+  if (xcrt_list == NULL)
     {
       gnutls_assert ();
+      ret = GNUTLS_E_MEMORY_ERROR;
       goto cleanup;
     }
 
@@ -703,7 +652,7 @@ cleanup:
 }
 
 
-/* Reads a private key from a token.
+/* Reads a certificate key from a token.
  */
 static int
 read_cert_url (gnutls_certificate_credentials_t res, const char *url)
@@ -729,6 +678,9 @@ read_cert_url (gnutls_certificate_credentials_t res, const char *url)
       goto cleanup;
     }
 
+  if (res->pin.cb)
+    gnutls_x509_crt_set_pin_function(crt, res->pin.cb, res->pin.data);
+
   ret = gnutls_x509_crt_import_pkcs11_url (crt, url, 0);
   if (ret == GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE)
     ret =
@@ -820,12 +772,10 @@ read_key_file (gnutls_certificate_credentials_t res,
   size_t size;
   char *data;
 
-#ifdef ENABLE_PKCS11
-  if (strncmp (keyfile, "pkcs11:", 7) == 0)
+  if (gnutls_url_is_supported(keyfile))
     {
       return read_key_url (res, keyfile);
     }
-#endif /* ENABLE_PKCS11 */
 
   data = read_binary_file (keyfile, &size);
 
@@ -1003,17 +953,15 @@ certificate_credentials_append_pkey (gnutls_certificate_credentials_t res,
  * @res: is a #gnutls_certificate_credentials_t structure.
  * @cert_list: contains a certificate list (path) for the specified private key
  * @cert_list_size: holds the size of the certificate list
- * @key: is a gnutls_x509_privkey_t key
+ * @key: is a #gnutls_x509_privkey_t key
  *
  * This function sets a certificate/private key pair in the
  * gnutls_certificate_credentials_t structure.  This function may be
  * called more than once, in case multiple keys/certificates exist for
- * the server.  For clients that wants to send more than its own end
+ * the server.  For clients that wants to send more than their own end
  * entity certificate (e.g., also an intermediate CA cert) then put
  * the certificate chain in @cert_list.
  *
- * 
- *
  * Returns: %GNUTLS_E_SUCCESS (0) on success, or a negative error code.
  *
  * Since: 2.4.0
@@ -1040,6 +988,9 @@ gnutls_certificate_set_x509_key (gnutls_certificate_credentials_t res,
       return ret;
     }
 
+  if (res->pin.cb)
+    gnutls_privkey_set_pin_function(pkey, res->pin.cb, res->pin.data);
+
   ret = gnutls_privkey_import_x509 (pkey, key, GNUTLS_PRIVKEY_IMPORT_COPY);
   if (ret < 0)
     {
@@ -1105,7 +1056,7 @@ cleanup:
  * @names_size: holds the size of the names list
  * @pcert_list: contains a certificate list (path) for the specified private key
  * @pcert_list_size: holds the size of the certificate list
- * @key: is a gnutls_x509_privkey_t key
+ * @key: is a #gnutls_privkey_t key
  *
  * This function sets a certificate/private key pair in the
  * gnutls_certificate_credentials_t structure.  This function may be
@@ -1114,8 +1065,8 @@ cleanup:
  * entity certificate (e.g., also an intermediate CA cert) then put
  * the certificate chain in @pcert_list. The @pcert_list and @key will
  * become part of the credentials structure and must not
- * be deallocated. They will be automatically deallocated when
- * @res is deinitialized.
+ * be deallocated. They will be automatically deallocated when @res
+ * is deinitialized.
  *
  * Returns: %GNUTLS_E_SUCCESS (0) on success, or a negative error code.
  *
@@ -1147,6 +1098,9 @@ gnutls_certificate_set_key (gnutls_certificate_credentials_t res,
         }
     }
 
+  if (res->pin.cb)
+    gnutls_privkey_set_pin_function(key, res->pin.cb, res->pin.data);
+
   ret = certificate_credentials_append_pkey (res, key);
   if (ret < 0)
     {
@@ -1191,8 +1145,9 @@ cleanup:
  * entity certificate, e.g., also an intermediate CA cert, then the
  * @certfile must contain the ordered certificate chain.
  *
- * This function can also accept PKCS #11 URLs at @keyfile and @certfile. In that case it
- * will import the private key and certificate indicated by the URLs.
+ * This function can also accept URLs at @keyfile and @certfile. In that case it
+ * will import the private key and certificate indicated by the URLs. Note
+ * that the supported URLs are the ones indicated by gnutls_url_is_supported().
  *
  * Returns: %GNUTLS_E_SUCCESS (0) on success, or a negative error code.
  **/
@@ -1546,8 +1501,9 @@ cleanup:
  * the client if a certificate request is sent. This can be disabled
  * using gnutls_certificate_send_x509_rdn_sequence().
  *
- * This function can also accept PKCS #11 URLs. In that case it
- * will import all certificates that are marked as trusted.
+ * This function can also accept URLs. In that case it
+ * will import all certificates that are marked as trusted. Note
+ * that the supported URLs are the ones indicated by gnutls_url_is_supported().
  *
  * Returns: number of certificates processed, or a negative error code on
  * error.
@@ -1590,110 +1546,6 @@ gnutls_certificate_set_x509_trust_file (gnutls_certificate_credentials_t cred,
   return ret;
 }
 
-#ifdef _WIN32
-static int
-set_x509_system_trust_file (gnutls_certificate_credentials_t cred)
-{
-HCERTSTORE store = CertOpenSystemStore(0, "CA");
-const CERT_CONTEXT *cert;
-const CRL_CONTEXT *crl;
-gnutls_datum_t data;
-int ret = 0;
-unsigned int i;
-
-  for (i=0;i<2;i++)
-    {
-    
-      if (i==0) store = CertOpenSystemStore(0, "ROOT");
-      else store = CertOpenSystemStore(0, "CA");
-    
-      if (store == NULL) return GNUTLS_E_FILE_ERROR;
-
-      cert = CertEnumCertificatesInStore(store, NULL);
-      crl = CertEnumCRLsInStore(store, NULL);
-
-      while(cert != NULL) 
-        {
-          if (cert->dwCertEncodingType == X509_ASN_ENCODING)
-            {
-              data.data = cert->pbCertEncoded;
-              data.size = cert->cbCertEncoded;
-              if (gnutls_certificate_set_x509_trust_mem (cred, &data, GNUTLS_X509_FMT_DER) > 0)
-                ret++;
-            }
-          cert = CertEnumCertificatesInStore(store, cert);
-        }
-
-      while(crl != NULL) 
-        {
-          if (crl->dwCertEncodingType == X509_ASN_ENCODING)
-            {
-              data.data = crl->pbCrlEncoded;
-              data.size = crl->cbCrlEncoded;
-            
-              gnutls_certificate_set_x509_crl_mem(cred, &data, GNUTLS_X509_FMT_DER);
-            }
-          crl = CertEnumCRLsInStore(store, crl);
-        }
-      CertCloseStore(store, 0);
-    }
-
-  return ret;
-}
-#elif defined(DEFAULT_TRUST_STORE_FILE)
-static int
-set_x509_system_trust_file (gnutls_certificate_credentials_t cred)
-{
-  int ret, r;
-  gnutls_datum_t cas;
-  size_t size;
-
-  cas.data = (void*)read_binary_file (DEFAULT_TRUST_STORE_FILE, &size);
-  if (cas.data == NULL)
-    {
-      gnutls_assert ();
-      return GNUTLS_E_FILE_ERROR;
-    }
-
-  cas.size = size;
-
-  ret = gnutls_certificate_set_x509_trust_mem(cred, &cas, GNUTLS_X509_FMT_PEM);
-
-  free (cas.data);
-
-  if (ret < 0)
-    {
-      gnutls_assert ();
-      return ret;
-    }
-  
-  r = ret;
-
-#ifdef DEFAULT_CRL_FILE
-  cas.data = (void*)read_binary_file (DEFAULT_CRL_FILE, &size);
-  if (cas.data == NULL)
-    {
-      gnutls_assert ();
-      return r;
-    }
-
-  cas.size = size;
-
-  ret = gnutls_certificate_set_x509_crl_mem(cred, &cas, GNUTLS_X509_FMT_PEM);
-
-  free (cas.data);
-
-  if (ret < 0)
-    {
-      gnutls_assert ();
-      return ret;
-    }
-#endif
-
-  return r;
-}
-#endif
-
 /**
  * gnutls_certificate_set_x509_system_trust:
  * @cred: is a #gnutls_certificate_credentials_t structure.
@@ -1712,25 +1564,7 @@ set_x509_system_trust_file (gnutls_certificate_credentials_t cred)
 int
 gnutls_certificate_set_x509_system_trust (gnutls_certificate_credentials_t cred)
 {
-#if !defined(_WIN32) && !defined(DEFAULT_TRUST_STORE_PKCS11) && !defined(DEFAULT_TRUST_STORE_FILE)
-  int r = GNUTLS_E_UNIMPLEMENTED_FEATURE;
-#else
-  int ret, r = 0;
-#endif
-
-#if defined(ENABLE_PKCS11) && defined(DEFAULT_TRUST_STORE_PKCS11)
-  ret = read_cas_url (cred, DEFAULT_TRUST_STORE_PKCS11);
-  if (ret > 0)
-    r += ret;
-#endif
-
-#ifdef DEFAULT_TRUST_STORE_FILE
-  ret = set_x509_system_trust_file(cred);
-  if (ret > 0)
-    r += ret;
-#endif
-
-  return r;
+  return gnutls_x509_trust_list_add_system_trust(cred->tlist, 0, 0);
 }
 
 static int
@@ -1956,306 +1790,6 @@ gnutls_certificate_set_x509_crl_file (gnutls_certificate_credentials_t res,
 
 #include <gnutls/pkcs12.h>
 
-static int
-parse_pkcs12 (gnutls_certificate_credentials_t res,
-              gnutls_pkcs12_t p12,
-              const char *password,
-              gnutls_x509_privkey_t * key,
-              gnutls_x509_crt_t * cert, gnutls_x509_crl_t * crl)
-{
-  gnutls_pkcs12_bag_t bag = NULL;
-  int idx = 0;
-  int ret;
-  size_t cert_id_size = 0;
-  size_t key_id_size = 0;
-  uint8_t cert_id[20];
-  uint8_t key_id[20];
-  int privkey_ok = 0;
-
-  *cert = NULL;
-  *key = NULL;
-  *crl = NULL;
-
-  /* find the first private key */
-  for (;;)
-    {
-      int elements_in_bag;
-      int i;
-
-      ret = gnutls_pkcs12_bag_init (&bag);
-      if (ret < 0)
-        {
-          bag = NULL;
-          gnutls_assert ();
-          goto done;
-        }
-
-      ret = gnutls_pkcs12_get_bag (p12, idx, bag);
-      if (ret == GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE)
-        break;
-      if (ret < 0)
-        {
-          gnutls_assert ();
-          goto done;
-        }
-
-      ret = gnutls_pkcs12_bag_get_type (bag, 0);
-      if (ret < 0)
-        {
-          gnutls_assert ();
-          goto done;
-        }
-
-      if (ret == GNUTLS_BAG_ENCRYPTED)
-        {
-          ret = gnutls_pkcs12_bag_decrypt (bag, password);
-          if (ret < 0)
-            {
-              gnutls_assert ();
-              goto done;
-            }
-        }
-
-      elements_in_bag = gnutls_pkcs12_bag_get_count (bag);
-      if (elements_in_bag < 0)
-        {
-          gnutls_assert ();
-          goto done;
-        }
-
-      for (i = 0; i < elements_in_bag; i++)
-        {
-          int type;
-          gnutls_datum_t data;
-
-          type = gnutls_pkcs12_bag_get_type (bag, i);
-          if (type < 0)
-            {
-              gnutls_assert ();
-              goto done;
-            }
-
-          ret = gnutls_pkcs12_bag_get_data (bag, i, &data);
-          if (ret < 0)
-            {
-              gnutls_assert ();
-              goto done;
-            }
-
-          switch (type)
-            {
-            case GNUTLS_BAG_PKCS8_ENCRYPTED_KEY:
-            case GNUTLS_BAG_PKCS8_KEY:
-              if (*key != NULL) /* too simple to continue */
-                {
-                  gnutls_assert ();
-                  break;
-                }
-
-              ret = gnutls_x509_privkey_init (key);
-              if (ret < 0)
-                {
-                  gnutls_assert ();
-                  goto done;
-                }
-
-              ret = gnutls_x509_privkey_import_pkcs8
-                (*key, &data, GNUTLS_X509_FMT_DER, password,
-                 type == GNUTLS_BAG_PKCS8_KEY ? GNUTLS_PKCS_PLAIN : 0);
-              if (ret < 0)
-                {
-                  gnutls_assert ();
-                  gnutls_x509_privkey_deinit (*key);
-                  goto done;
-                }
-
-              key_id_size = sizeof (key_id);
-              ret =
-                gnutls_x509_privkey_get_key_id (*key, 0, key_id,
-                                                &key_id_size);
-              if (ret < 0)
-                {
-                  gnutls_assert ();
-                  gnutls_x509_privkey_deinit (*key);
-                  goto done;
-                }
-
-              privkey_ok = 1;   /* break */
-              break;
-            default:
-              break;
-            }
-        }
-
-      idx++;
-      gnutls_pkcs12_bag_deinit (bag);
-
-      if (privkey_ok != 0)      /* private key was found */
-        break;
-    }
-
-  if (privkey_ok == 0)          /* no private key */
-    {
-      gnutls_assert ();
-      return GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE;
-    }
-
-  /* now find the corresponding certificate 
-   */
-  idx = 0;
-  bag = NULL;
-  for (;;)
-    {
-      int elements_in_bag;
-      int i;
-
-      ret = gnutls_pkcs12_bag_init (&bag);
-      if (ret < 0)
-        {
-          bag = NULL;
-          gnutls_assert ();
-          goto done;
-        }
-
-      ret = gnutls_pkcs12_get_bag (p12, idx, bag);
-      if (ret == GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE)
-        break;
-      if (ret < 0)
-        {
-          gnutls_assert ();
-          goto done;
-        }
-
-      ret = gnutls_pkcs12_bag_get_type (bag, 0);
-      if (ret < 0)
-        {
-          gnutls_assert ();
-          goto done;
-        }
-
-      if (ret == GNUTLS_BAG_ENCRYPTED)
-        {
-          ret = gnutls_pkcs12_bag_decrypt (bag, password);
-          if (ret < 0)
-            {
-              gnutls_assert ();
-              goto done;
-            }
-        }
-
-      elements_in_bag = gnutls_pkcs12_bag_get_count (bag);
-      if (elements_in_bag < 0)
-        {
-          gnutls_assert ();
-          goto done;
-        }
-
-      for (i = 0; i < elements_in_bag; i++)
-        {
-          int type;
-          gnutls_datum_t data;
-
-          type = gnutls_pkcs12_bag_get_type (bag, i);
-          if (type < 0)
-            {
-              gnutls_assert ();
-              goto done;
-            }
-
-          ret = gnutls_pkcs12_bag_get_data (bag, i, &data);
-          if (ret < 0)
-            {
-              gnutls_assert ();
-              goto done;
-            }
-
-          switch (type)
-            {
-            case GNUTLS_BAG_CERTIFICATE:
-              if (*cert != NULL)        /* no need to set it again */
-                {
-                  gnutls_assert ();
-                  break;
-                }
-
-              ret = gnutls_x509_crt_init (cert);
-              if (ret < 0)
-                {
-                  gnutls_assert ();
-                  goto done;
-                }
-
-              ret =
-                gnutls_x509_crt_import (*cert, &data, GNUTLS_X509_FMT_DER);
-              if (ret < 0)
-                {
-                  gnutls_assert ();
-                  gnutls_x509_crt_deinit (*cert);
-                  goto done;
-                }
-
-              /* check if the key id match */
-              cert_id_size = sizeof (cert_id);
-              ret =
-                gnutls_x509_crt_get_key_id (*cert, 0, cert_id, &cert_id_size);
-              if (ret < 0)
-                {
-                  gnutls_assert ();
-                  gnutls_x509_crt_deinit (*cert);
-                  goto done;
-                }
-
-              if (memcmp (cert_id, key_id, cert_id_size) != 0)
-                {               /* they don't match - skip the certificate */
-                  gnutls_x509_crt_deinit (*cert);
-                  *cert = NULL;
-                }
-              break;
-
-            case GNUTLS_BAG_CRL:
-              if (*crl != NULL)
-                {
-                  gnutls_assert ();
-                  break;
-                }
-
-              ret = gnutls_x509_crl_init (crl);
-              if (ret < 0)
-                {
-                  gnutls_assert ();
-                  goto done;
-                }
-
-              ret = gnutls_x509_crl_import (*crl, &data, GNUTLS_X509_FMT_DER);
-              if (ret < 0)
-                {
-                  gnutls_assert ();
-                  gnutls_x509_crl_deinit (*crl);
-                  goto done;
-                }
-              break;
-
-            case GNUTLS_BAG_ENCRYPTED:
-              /* XXX Bother to recurse one level down?  Unlikely to
-                 use the same password anyway. */
-            case GNUTLS_BAG_EMPTY:
-            default:
-              break;
-            }
-        }
-
-      idx++;
-      gnutls_pkcs12_bag_deinit (bag);
-    }
-
-  ret = 0;
-
-done:
-  if (bag)
-    gnutls_pkcs12_bag_deinit (bag);
-
-  return ret;
-}
 
 /**
  * gnutls_certificate_set_x509_simple_pkcs12_file:
@@ -2323,8 +1857,7 @@ int
  * be called more than once (in case multiple keys/certificates exist
  * for the server).
  *
- * MAC:ed PKCS#12 files are supported.  Encrypted PKCS#12 bags are
- * supported.  Encrypted PKCS#8 private keys are supported.  However,
+ * Encrypted PKCS#12 bags and PKCS#8 private keys are supported.  However,
  * only password based security, and the same password for all
  * operations, are supported.
  *
@@ -2349,8 +1882,9 @@ int
 {
   gnutls_pkcs12_t p12;
   gnutls_x509_privkey_t key = NULL;
-  gnutls_x509_crt_t cert = NULL;
+  gnutls_x509_crt_t *chain = NULL;
   gnutls_x509_crl_t crl = NULL;
+  unsigned int chain_size = 0, i;
   int ret;
 
   ret = gnutls_pkcs12_init (&p12);
@@ -2379,7 +1913,8 @@ int
         }
     }
 
-  ret = parse_pkcs12 (res, p12, password, &key, &cert, &crl);
+  ret = gnutls_pkcs12_simple_parse (p12, password, &key, &chain, &chain_size, 
+                             NULL, NULL, &crl, 0);
   gnutls_pkcs12_deinit (p12);
   if (ret < 0)
     {
@@ -2387,15 +1922,21 @@ int
       return ret;
     }
 
-  if (key && cert)
+  if (key && chain)
     {
-      ret = gnutls_certificate_set_x509_key (res, &cert, 1, key);
+      ret = gnutls_certificate_set_x509_key (res, chain, chain_size, key);
       if (ret < 0)
         {
           gnutls_assert ();
           goto done;
         }
     }
+  else
+    {
+      gnutls_assert();
+      ret = GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE;
+      goto done;
+    }
 
   if (crl)
     {
@@ -2410,8 +1951,12 @@ int
   ret = 0;
 
 done:
-  if (cert)
-    gnutls_x509_crt_deinit (cert);
+  if (chain)
+    {
+      for (i=0;i<chain_size;i++)
+        gnutls_x509_crt_deinit (chain[i]);
+      gnutls_free(chain);
+    }
   if (key)
     gnutls_x509_privkey_deinit (key);
   if (crl)
@@ -2435,3 +1980,25 @@ gnutls_certificate_free_crls (gnutls_certificate_credentials_t sc)
   /* do nothing for now */
   return;
 }
+
+/**
+ * gnutls_certificate_credentials_t:
+ * @cred: is a #gnutls_certificate_credentials_t structure.
+ * @fn: A PIN callback
+ * @userdata: Data to be passed in the callback
+ *
+ * This function will set a callback function to be used when
+ * required to access a protected object. This function overrides any other
+ * global PIN functions.
+ *
+ * Note that this function must be called right after initialization
+ * to have effect.
+ *
+ * Since: 3.1.0
+ **/
+void gnutls_certificate_set_pin_function (gnutls_certificate_credentials_t cred,
+                                          gnutls_pin_callback_t fn, void *userdata)
+{
+  cred->pin.cb = fn;
+  cred->pin.data = userdata;
+}
