@@ -103,28 +103,6 @@ static int do_handshake (socket_st * socket);
 static void init_global_tls_stuff (void);
 static int cert_verify_ocsp (gnutls_session_t session);
 
-/* Helper functions to load a certificate and key
- * files into memory.
- */
-static gnutls_datum_t
-load_file (const char *file)
-{
-  gnutls_datum_t loaded_file = { NULL, 0 };
-  size_t length;
-
-  loaded_file.data = (void*)read_binary_file (file, &length);
-  if (loaded_file.data)
-    loaded_file.size = (unsigned int) length;
-
-  return loaded_file;
-}
-
-static void
-unload_file (gnutls_datum_t* data)
-{
-  free (data->data);
-}
-
 #define MAX_CRT 6
 static unsigned int x509_crt_size;
 static gnutls_pcert_st x509_crt[MAX_CRT];
@@ -164,10 +142,6 @@ load_keys (void)
   unsigned int i;
   gnutls_datum_t data = { NULL, 0 };
   gnutls_x509_crt_t crt_list[MAX_CRT];
-#ifdef ENABLE_PKCS11
-  gnutls_pkcs11_privkey_t pkcs11_key;
-#endif
-  gnutls_x509_privkey_t tmp_key;
   unsigned char keyid[GNUTLS_OPENPGP_KEYID_SIZE];
 
   if (x509_certfile != NULL && x509_keyfile != NULL)
@@ -177,6 +151,7 @@ load_keys (void)
         {
           crt_num = 1;
           gnutls_x509_crt_init (&crt_list[0]);
+          gnutls_x509_crt_set_pin_function(crt_list[0], pin_callback, NULL);
 
           ret =
             gnutls_x509_crt_import_pkcs11_url (crt_list[0], x509_certfile, 0);
@@ -197,8 +172,8 @@ load_keys (void)
 #endif /* ENABLE_PKCS11 */
         {
 
-          data = load_file (x509_certfile);
-          if (data.data == NULL)
+          ret = gnutls_load_file (x509_certfile, &data);
+          if (ret < 0)
             {
               fprintf (stderr, "*** Error loading cert file.\n");
               exit (1);
@@ -241,7 +216,7 @@ load_keys (void)
           gnutls_x509_crt_deinit(crt_list[i]);
         }
 
-      unload_file (&data);
+      gnutls_free (data.data);
 
       ret = gnutls_privkey_init(&x509_key);
       if (ret < 0)
@@ -251,21 +226,12 @@ load_keys (void)
            exit (1);
          }
 
-#ifdef ENABLE_PKCS11
-      if (strncmp (x509_keyfile, "pkcs11:", 7) == 0)
+      gnutls_privkey_set_pin_function(x509_key, pin_callback, NULL);
+
+      if (gnutls_url_is_supported(x509_keyfile) != 0)
         {
-          gnutls_pkcs11_privkey_init (&pkcs11_key);
-
           ret =
-            gnutls_pkcs11_privkey_import_url (pkcs11_key, x509_keyfile, 0);
-          if (ret < 0)
-            {
-              fprintf (stderr, "*** Error loading url: %s\n",
-                       gnutls_strerror (ret));
-              exit (1);
-            }
-
-          ret = gnutls_privkey_import_pkcs11( x509_key, pkcs11_key, GNUTLS_PRIVKEY_IMPORT_AUTO_RELEASE);
+            gnutls_privkey_import_url (x509_key, x509_keyfile, 0);
           if (ret < 0)
             {
               fprintf (stderr, "*** Error loading url: %s\n",
@@ -274,27 +240,15 @@ load_keys (void)
             }
         }
       else
-#endif /* ENABLE_PKCS11 */
         {
-          data = load_file (x509_keyfile);
-          if (data.data == NULL)
+          ret = gnutls_load_file (x509_keyfile, &data);
+          if (ret < 0)
             {
               fprintf (stderr, "*** Error loading key file.\n");
               exit (1);
             }
 
-          gnutls_x509_privkey_init (&tmp_key);
-
-          ret =
-            gnutls_x509_privkey_import (tmp_key, &data, x509ctype);
-          if (ret < 0)
-            {
-              fprintf (stderr, "*** Error loading key file: %s\n",
-                       gnutls_strerror (ret));
-              exit (1);
-            }
-
-          ret = gnutls_privkey_import_x509( x509_key, tmp_key, GNUTLS_PRIVKEY_IMPORT_AUTO_RELEASE);
+          ret = gnutls_privkey_import_x509_raw( x509_key, &data, x509ctype, NULL, 0);
           if (ret < 0)
             {
               fprintf (stderr, "*** Error loading url: %s\n",
@@ -302,7 +256,7 @@ load_keys (void)
               exit (1);
             }
 
-          unload_file (&data);
+          gnutls_free(data.data);
         }
 
       fprintf (stdout, "Processed %d client X.509 certificates...\n",
@@ -320,8 +274,8 @@ load_keys (void)
     {
       gnutls_openpgp_crt_t tmp_pgp_crt;
 
-      data = load_file (pgp_certfile);
-      if (data.data == NULL)
+      ret = gnutls_load_file (pgp_certfile, &data);
+      if (ret < 0)
         {
           fprintf (stderr, "*** Error loading PGP cert file.\n");
           exit (1);
@@ -339,7 +293,7 @@ load_keys (void)
           exit (1);
         }
  
-      unload_file (&data);
+      gnutls_free (data.data);
 
       ret = gnutls_privkey_init(&pgp_key);
       if (ret < 0)
@@ -349,20 +303,11 @@ load_keys (void)
            exit (1);
          }
 
-#ifdef ENABLE_PKCS11
-      if (strncmp (pgp_keyfile, "pkcs11:", 7) == 0)
+      gnutls_privkey_set_pin_function(pgp_key, pin_callback, NULL);
+
+      if (gnutls_url_is_supported (pgp_keyfile))
         {
-          gnutls_pkcs11_privkey_init (&pkcs11_key);
-
-          ret = gnutls_pkcs11_privkey_import_url (pkcs11_key, pgp_keyfile, 0);
-          if (ret < 0)
-            {
-              fprintf (stderr, "*** Error loading url: %s\n",
-                       gnutls_strerror (ret));
-              exit (1);
-            }
-
-          ret = gnutls_privkey_import_pkcs11( pgp_key, pkcs11_key, GNUTLS_PRIVKEY_IMPORT_AUTO_RELEASE);
+          ret = gnutls_privkey_import_url( pgp_key, pgp_keyfile, 0);
           if (ret < 0)
             {
               fprintf (stderr, "*** Error loading url: %s\n",
@@ -371,45 +316,18 @@ load_keys (void)
             }
         }
       else
-#endif /* ENABLE_PKCS11 */
         {
-          gnutls_openpgp_privkey_t tmp_pgp_key;
-
-          data = load_file (pgp_keyfile);
-          if (data.data == NULL)
-            {
-              fprintf (stderr, "*** Error loading PGP key file.\n");
-              exit (1);
-            }
-
-          gnutls_openpgp_privkey_init (&tmp_pgp_key);
-
-          ret =
-            gnutls_openpgp_privkey_import (tmp_pgp_key, &data,
-                                           GNUTLS_OPENPGP_FMT_BASE64, NULL,
-                                           0);
+          ret = gnutls_load_file (pgp_keyfile, &data);
           if (ret < 0)
             {
-              fprintf (stderr,
-                       "*** Error loading PGP key file: %s\n",
-                       gnutls_strerror (ret));
+              fprintf (stderr, "*** Error loading key file.\n");
               exit (1);
             }
 
           if (HAVE_OPT(PGPSUBKEY))
-            {
-              ret =
-                gnutls_openpgp_privkey_set_preferred_key_id (tmp_pgp_key, keyid);
-              if (ret < 0)
-                {
-                  fprintf (stderr,
-                      "*** Error setting preferred sub key id (%s): %s\n",
-                      OPT_ARG(PGPSUBKEY), gnutls_strerror (ret));
-                  exit (1);
-                }
-            }
-
-          ret = gnutls_privkey_import_openpgp( pgp_key, tmp_pgp_key, GNUTLS_PRIVKEY_IMPORT_AUTO_RELEASE);
+            ret = gnutls_privkey_import_openpgp_raw( pgp_key, &data, x509ctype, keyid, NULL);
+          else
+            ret = gnutls_privkey_import_openpgp_raw( pgp_key, &data, x509ctype, NULL, NULL);
           if (ret < 0)
             {
               fprintf (stderr, "*** Error loading url: %s\n",
@@ -417,7 +335,7 @@ load_keys (void)
               exit (1);
             }
 
-          unload_file (&data);
+          gnutls_free(data.data);
         }
 
 
@@ -674,7 +592,8 @@ init_tls_session (const char *hostname)
                                 strlen (hostname));
     }
 
-  gnutls_dh_set_prime_bits (session, 512);
+  if (HAVE_OPT(DH_BITS)) 
+    gnutls_dh_set_prime_bits( session, OPT_VALUE_DH_BITS);
 
   gnutls_credentials_set (session, GNUTLS_CRD_ANON, anon_cred);
   if (srp_cred)
@@ -685,6 +604,7 @@ init_tls_session (const char *hostname)
 
   gnutls_certificate_set_retrieve_function2 (xcred, cert_callback);
   gnutls_certificate_set_verify_function (xcred, cert_verify_callback);
+  gnutls_certificate_set_verify_flags (xcred, 0);
 
   /* send the fingerprint */
 #ifdef ENABLE_OPENPGP
@@ -856,7 +776,7 @@ main (int argc, char **argv)
     }
 
 #ifdef ENABLE_PKCS11
-  pkcs11_common ();
+//  pkcs11_common ();
 #endif
 
   if (hostname == NULL)
@@ -963,7 +883,7 @@ after_handshake:
   fflush (stderr);
 
   /* do not buffer */
-#if !(defined _WIN32 || defined __WIN32__)
+#ifndef _WIN32
   setbuf (stdin, NULL);
 #endif
   setbuf (stdout, NULL);
@@ -1122,9 +1042,15 @@ const char* rest = NULL;
       exit(0);
     }
 
-  if (HAVE_OPT(BENCHMARK_TLS))
+  if (HAVE_OPT(BENCHMARK_TLS_CIPHERS))
     {
-      benchmark_tls(OPT_VALUE_DEBUG); 
+      benchmark_tls(OPT_VALUE_DEBUG, 1); 
+      exit(0);
+    }
+
+  if (HAVE_OPT(BENCHMARK_TLS_KX))
+    {
+      benchmark_tls(OPT_VALUE_DEBUG, 0); 
       exit(0);
     }
 
@@ -1220,19 +1146,6 @@ const char* rest = NULL;
     }
 }
 
-void cli_version (void);
-
-void
-cli_version (void)
-{
-  const char *p = PACKAGE_NAME;
-  if (strcmp (gnutls_check_version (NULL), PACKAGE_VERSION) != 0)
-    p = PACKAGE_STRING;
-  version_etc (stdout, program_name, p, gnutls_check_version (NULL),
-               "Nikos Mavrogiannopoulos", (char *) NULL);
-}
-
-
 static void
 check_rehandshake (socket_st * socket, int ret)
 {
@@ -1270,6 +1183,7 @@ do_handshake (socket_st * socket)
                             gl_fd_to_handle (socket->fd));
   do
     {
+      gnutls_handshake_set_timeout( socket->session, GNUTLS_DEFAULT_HANDSHAKE_TIMEOUT);
       ret = gnutls_handshake (socket->session);
 
       if (ret < 0)
@@ -1282,7 +1196,7 @@ do_handshake (socket_st * socket)
   if (ret == 0)
     {
       /* print some information */
-      print_info (socket->session, print_cert);
+      print_info (socket->session, print_cert, verbose);
       socket->secure = 1;
     }
   else
@@ -1404,6 +1318,7 @@ init_global_tls_stuff (void)
       fprintf (stderr, "Certificate allocation memory error\n");
       exit (1);
     }
+  gnutls_certificate_set_pin_function(xcred, pin_callback, NULL);
 
   if (x509_cafile != NULL)
     {
