@@ -730,17 +730,6 @@ listen_socket (const char *name, int listen_port, int socktype)
           continue;
         }
 
-#ifdef HAVE_IPV6
-      if (ptr->ai_family == AF_INET6)
-        {
-          yes = 1;
-          /* avoid listen on ipv6 addresses failing
-	   * because already listening on ipv4 addresses: */
-          setsockopt (s, IPPROTO_IPV6, IPV6_V6ONLY,
-                          (const void *) &yes, sizeof (yes));
-        }
-#endif
-
       if (socktype == SOCK_STREAM)
         {
           yes = 1;
@@ -1311,7 +1300,7 @@ static void tcp_server(const char* name, int port)
                                 human_addr ((struct sockaddr *)
                                             &client_address, calen, topbuf,
                                             sizeof (topbuf)));
-                        print_info (j->tls_session, verbose);
+                        print_info (j->tls_session, verbose, verbose);
                         if (gnutls_auth_get_type (j->tls_session) == GNUTLS_CRD_CERTIFICATE)
                           cert_verify(j->tls_session, NULL);
                       }
@@ -1436,7 +1425,7 @@ static void tcp_server(const char* name, int port)
                                             &client_address, calen, topbuf,
                                             sizeof (topbuf)));
 
-                        print_info (j->tls_session, verbose);
+                        print_info (j->tls_session, verbose, verbose);
                         if (gnutls_auth_get_type (j->tls_session) == GNUTLS_CRD_CERTIFICATE)
                           cert_verify(j->tls_session, NULL);
                       }
@@ -1622,18 +1611,6 @@ static void cmd_parser (int argc, char **argv)
 
 }
 
-extern void serv_version (void);
-
-void
-serv_version (void)
-{
-  const char *p = PACKAGE_NAME;
-  if (strcmp (gnutls_check_version (NULL), PACKAGE_VERSION) != 0)
-    p = PACKAGE_STRING;
-  version_etc (stdout, program_name, p, gnutls_check_version (NULL),
-               "Nikos Mavrogiannopoulos", (char *) NULL);
-}
-
 /* session resuming support */
 
 #define SESSION_ID_SIZE 32
